// Code adapted from: http://en.highscore.de/cpp/boost/parser.html

#include <boost/spirit/include/qi.hpp>
// Boost Fusion is a Boost library which enables
// the user to work with heterogeneous collections
// of data (i.e. tuple)
#include <boost/fusion/include/adapt_struct.hpp>

// Extends C/C++ unions. It defines a stack-based union
// container for heterogeneous types (not just C POD's
// like union, but also C++ user defined structures as
// well; also works with recursive data types (i.e. data
// types that have circular dependencies, like for example
// a recursively defined list))
#include <boost/variant.hpp>

// A Boost library which allows the user to use range based
// (and collection iterator based) for loops
#include <boost/foreach.hpp>

#include <string>
#include <vector>
#include <iostream>

using namespace boost::spirit;

struct json_object;
struct json_array;

// A JSON value can be of several types (numerical value,
// a string of characters, another more complex JSON object
// or an array of JSON values
typedef boost::variant<std::string, double,
        boost::recursive_wrapper<json_object>,
        boost::recursive_wrapper<json_array> > json_value;

// An array in JSON is defined as a vector of JSON values
struct json_array
{
    std::vector<json_value> array;
};

// macro-based code that creates a random access bidirectional
// sequence of heterogeneous elements
// In general such a structure is declared as
//  BOOST_FUSION_ADAPT_STRUCT(
//      struct_name,
//      (member_type0, member_name0)
//      (member_type1, member_name1)
//      ...
//  )
// allows (amortized) constant-time access to the elements
// of the sequence struct_name, see for example:
// http://www.boost.org/doc/libs/1_56_0/libs/fusion/doc/html/fusion/sequence/concepts/random_access_sequence.html
BOOST_FUSION_ADAPT_STRUCT(
    json_array,
        (std::vector<json_value>, array)
)

struct json_member
{
    std::string key;
    json_value value;
};

BOOST_FUSION_ADAPT_STRUCT(
    json_member,
        (std::string, key)
        (json_value, value)
)

struct json_object
{
    std::vector<json_member> members;
};

BOOST_FUSION_ADAPT_STRUCT(
    json_object,
        (std::vector<json_member>, members)
)


// the Skipper template argument is used by the parser rules
// to know which characters from the input stream it should
// ignore (in our case it will be ASCII space type characters)
template <typename Iterator, typename Skipper>
        struct json : qi::grammar<Iterator, json_object(), Skipper>
        {
            json() : json::base_type(object)
            {
                object %= "{" >> member % "," >> "}";
                member %= str >> ":" >> value;
                str %= qi::lexeme["\"" >> *~qi::lit('"') >> "\""];
                value %= str | number | object | array | qi::string("true") |
                        qi::string("false") | qi::string("null");
                number %= qi::double_;
                array %= "[" >> value % "," >> "]";

            }

            // the production rules of the grammar
            // the second template argument specifies the
            // actions to be executed on each rule (they are
            // defined in the next data structure)
            qi::rule<Iterator, json_object(), Skipper> object;
            qi::rule<Iterator, json_member(), Skipper> member;
            qi::rule<Iterator, std::string(), Skipper> str;
            qi::rule<Iterator, json_value(), Skipper> value;
            qi::rule<Iterator, double(), Skipper> number;
            qi::rule<Iterator, json_array(), Skipper> array;
        };

// the actions to be executed for each rule are specified
// by overloading operator() for each terminal and non-terminal
// This is one way of doing things; the next exercise shows how
// you can add semantic actions to rules in a simpler way
struct output
         :public boost::static_visitor<>

{
    void operator()(const std::string &s) const { std::cout << s << std::endl; }
    void operator()(double d) const {std::cout << d << std::endl; }
    void operator()(const json_object & o) const
    {
        BOOST_FOREACH(json_member member, o.members)
                    {
                        std::cout << member.key << std::endl;
                        boost::apply_visitor(output(), member.value);
                    }
    }
        void operator()(const json_array & a) const
    {
        BOOST_FOREACH(json_value value, a.array)
                    {
                        boost::apply_visitor(output(), value);
                    }
    }
};

int main()
{
    std::string s = "{\"Boost Spirit Parser\":{\"Language\":\"C++\",\"Non "
            "Terminals\":[\"object\",\"member\",\"str\",\"value\",\"number\",\"array\"]}}";
    // create an instance of the parser, which handles strings as inputs
    json<std::string::iterator, ascii::space_type> json_parser;
    json_object json_data;
    bool success = qi::phrase_parse(s.begin(), s.end(), json_parser, ascii::space,
            json_data);
    if (success)
    {
        BOOST_FOREACH(json_member member, json_data.members)
                    {
                        std::cout << member.key << std::endl;
                        boost::apply_visitor(output(), member.value);
                    }
    }
}
